/*
 * main.c
 *
 *  Created on: Oct 26, 2015
 *      Author: mbodis
 */
#define F_CPU 8000000UL
#include <avr/io.h>
#include <util/delay.h>

const int COL_MIN = 1;
const int COL_MAX = 5;
const int ROW_MIN = 1;
const int ROW_MAX = 4;

//------- snake
int BACKTRACK_MAX = 5+4; // depth search with limit
int SNAKE_MAX = 5 * 4;
int sX[5 * 4];
int sY[5 * 4];
int sL = 0;
int fX, fY; // food-[x,y]
int foodEaten = 0;
//-------

int main(void) {
	DDRB = 0xFF; /* set PB0 to output */
	DDRC = 0x7f; /* set PC0 to output */
	DDRD = 0xFF; /* set PD0 to output */

	while (1) {

		/* random int between 5 and 44 */
		int time = rand() % 40 + 5;

		switch(rand() % 8 ){// 0 - 7
		case 0:
			blinkAll(time);
			break;
		case 1:
			heart(time, 10);
			break;
		case 2:
			rainEffect(time, 0);
			break;
		case 3:
			rainEffect(time, 1);
			break;
		case 4:
			sideEffect(time, 0);
			break;
		case 5:
			sideEffect(time, 1);
			break;
		case 6:
			randomBlinking(time);
			break;
		case 7:
			snake();
			blinkAll(time);
			break;
		}

	}

	return 0;
}

/* ------------------------------------------------------------ */
/* SNAKE */
/* ------------------------------------------------------------ */

void snake() {

	initSnake();
	int time = 1;
	int blink = 0;
	int live = 1;
	while (live == 1) {

		if (blink % 5 == 0) {

			int hasMoved = 0;
			while (hasMoved == 0) {

				switch (chooseNextMoveBacktrack()) { // direction 1-4
				case 1: // left
					hasMoved = moveSnakeLeft();
					break;
				case 2: // right
					hasMoved = moveSnakeRight();
					break;
				case 3: // up
					hasMoved = moveSnakeUp();
					break;
				case 4: // down
					hasMoved = moveSnakeDown();
					break;
				}
				//end game test
				if (hasSnakeOptions() == 0) {
					live = 0;
					break;
				}
			}
			hasSnakeEatFood();
		}
		showSnake(time);

		if (blink % 3 == 0)
			lightOn(fX, fY, time);

		blink++;
		if (blink == 30)
			blink = 0;
	}
}

void initSnake() {
	for (int i = 0; i < SNAKE_MAX; i++) {
		sX[i] = 0;
		sY[i] = 0;
	}
	sX[0] = 2;
	sY[0] = 1;
	sX[1] = 1;
	sY[1] = 1;
	sL = 2;
	generateFood();
}

void generateFood() {
	int foodPlaced = 0;

	while (foodPlaced == 0) {
		fX = rand() % COL_MAX + 1;
		fY = rand() % ROW_MAX + 1;
		if (isSnakeOnPosition(fX, fY) == 0) {
			foodPlaced = 1;
		}
	}
}

void showSnake(int time) {
	for (int i = 0; i < SNAKE_MAX; i++) {
		lightOn(sX[i], sY[i], time);
	}
}

int moveSnakeLeft() {
	if (sX[0] > COL_MIN && isSnakeOnPosition(sX[0] - 1, sY[0]) == 0) {
		moveSnakeToPosition(sX[0] - 1, sY[0]);
		return 1;
	}

	return 0;
}

int moveSnakeRight() {
	if (sX[0] < COL_MAX && isSnakeOnPosition(sX[0] + 1, sY[0]) == 0) {
		moveSnakeToPosition(sX[0] + 1, sY[0]);
		return 1;
	}
	return 0;
}
int moveSnakeUp() {
	if (sY[0] < ROW_MAX && isSnakeOnPosition(sX[0], sY[0] + 1) == 0) {
		moveSnakeToPosition(sX[0], sY[0] + 1);
		return 1;
	}
	return 0;
}
int moveSnakeDown() {
	if (sY[0] > ROW_MIN && isSnakeOnPosition(sX[0], sY[0] - 1) == 0) {
		moveSnakeToPosition(sX[0], sY[0] - 1);
		return 1;
	}
	return 0;
}

int isSnakeOnPosition(int x, int y) {
	for (int i = 0; i < sL; i++) {
		if (sX[i] == x && sY[i] == y) {
			return 1;
		}
	}
	return 0;
}

void moveSnakeToPosition(int x, int y) {

	if (foodEaten == 2) {
		foodEaten=0;
		sL++;
	}
	for (int i = sL - 1; i >= 0; i--) {
		if (i == 0) {
			sX[i] = x;
			sY[i] = y;

		} else {
			sX[i] = sX[i - 1];
			sY[i] = sY[i - 1];
		}

	}
}

void hasSnakeEatFood() {
	for (int i = 0; i < sL; i++) {
		if (sX[i] == fX && sY[i] == fY) {
			foodEaten++;
			generateFood();
		}
	}
}

int hasSnakeOptions() {
	//left
	if (sX[0] > COL_MIN && isSnakeOnPosition(sX[0] - 1, sY[0]) == 0) {
		return 1;
	}
	// right
	if (sX[0] < COL_MAX && isSnakeOnPosition(sX[0] + 1, sY[0]) == 0) {
		return 1;
	}
	// up
	if (sY[0] < ROW_MAX && isSnakeOnPosition(sX[0], sY[0] + 1) == 0) {
		return 1;
	}
	// down
	if (sY[0] > ROW_MIN && isSnakeOnPosition(sX[0], sY[0] - 1) == 0) {
		return 1;
	}

	return 0;
}

/*
 * 1: left
 * 2: right
 * 3: up
 * 4: down
 */
int chooseNextMoveRandom(){
	return rand() % 4 + 1; // <1-4>
}

/*
 * 1: left
 * 2: right
 * 3: up
 * 4: down
 */
int chooseNextMoveBacktrack(){
	int xx[COL_MAX];
	int yy[ROW_MAX];

	// maximu possible moves
	for (int i=0; i< BACKTRACK_MAX; i++){
		if (snakeBacktrack(sX[0]-1, sY[0], i) == 1){
			return 1;
		}
		if (snakeBacktrack(sX[0]+1, sY[0], i) == 1){
			// next move sX[0]+1, sY[0]
			return 2;
		}
		if (snakeBacktrack(sX[0], sY[0]+1, i) == 1){
			// next move sX[0], sY[0]+1
			return 3;
		}
		if (snakeBacktrack(sX[0], sY[0]-1, i) == 1){
			// next move sX[0], sY[0]-1
			return 4;
		}
	}

	// fallback
	return chooseNextMoveRandom();
}

/*
 * 0 - not found
 * 1 - found
 */
int snakeBacktrack(int posX, int posY, int level){
	// out of field
	if (posX<COL_MIN || posX>COL_MAX || posY<ROW_MIN || posY>ROW_MAX){
		return 0;
	}

	// snake body
	if (isSnakeOnPosition(posX, posY) == 1){
		return 0;
	}

	if (level == 0){
		if (posX == fX && posY == fY){
			return 1;
		}else{
			return 0;
		}
	}else if (level > 0){
		// left
		if (snakeBacktrack(posX-1, posY, level-1) == 1){
			return 1;
		}
		// right
		if (snakeBacktrack(posX+1, posY, level-1) == 1){
			return 1;
		}
		// up
		if (snakeBacktrack(posX, posY+1, level-1) == 1){
			return 1;
		}
		// down
		if (snakeBacktrack(posX, posY-1, level-1) == 1){
			return 1;
		}
	}

	return 0;
}

/* ------------------------------------------------------------ */
/* SCREEN EFFECTS  */
/* ------------------------------------------------------------ */

void randomBlinking(int time) {
	int repeat = rand() % 40 + 5;
	while (repeat > 1) {
		int x = rand() % COL_MAX + 1;
		int y = rand() % ROW_MAX + 1;
		lightOn(x, y, time);
		repeat--;
	}
}

void sideEffect(int time, int direction) {
	int repeat = rand() % 40 + 5;
	while (repeat > 1) {
		int row = rand() % 4 + 1; // 1-4
		if (direction == 1) {
			sideEffectLeftToRight(row, time);
		} else {
			sideEffectRightToLeft(row, time);
		}
		repeat--;
	}
}

/*
 * col 1-5 column
 * time = sleep time miliseconds
 */
void sideEffectLeftToRight(int row, int time) {
	for (int i = 1; i <= 5; i++) {
		lightOn(i, row, time);
	}
}
void sideEffectRightToLeft(int row, int time) {
	for (int i = 5; i <= 1; i--) {
		lightOn(i, row, time);
	}
}

void rainEffect(int time, int direct) {
	int repeat = rand() % 40 + 5;
	while (repeat > 1) {
		int col = rand() % 5 + 1; // 1-5
		if (direct == 1) {
			rainEffectCol(col, time);
		} else {
			rainEffectColUp(col, time);
		}
		repeat--;
	}

}

/*
 * col 1-5 column
 * time = sleep time miliseconds
 */
void rainEffectCol(int col, int time) {
	for (int i = 1; i <= 4; i++) {
		lightOn(col, i, time);
	}
}
void rainEffectColUp(int col, int time) {
	for (int i = 4; i <= 1; i--) {
		lightOn(col, i, time);
	}
}

void heart(int time, int repeat) {
	time = 10;
	int i = 0;

	while (i < repeat) {
		lightOn(3, 3, time);
		lightOn(2, 4, time);
		lightOn(1, 3, time);
		lightOn(2, 2, time);
		lightOn(3, 1, time);
		lightOn(4, 2, time);
		lightOn(5, 3, time);
		lightOn(4, 4  , time);
		i++;
	}
}

void blinkAll(int time) {
	for (int row = 1; row <= 4; row++) {
		for (int col = 1; col <= 5; col++) {
			lightOn(col, row, time);
		}
	}

}
/* ------------------------------------------------------------ */
/* COMMON */
/* ------------------------------------------------------------ */

/*
 * mapping leds to my current setup
 */
void lightOn(int x, int y, int sleepTime) {
	lightOff();
	if (x == 1 && y == 1)
		PORTD = 0x02; // bit-2 - D
	if (x == 1 && y == 2)
		PORTD = 0x08; // bit-4 - D
	if (x == 1 && y == 3)
		PORTB = 0x80; // bit-8 - B
	if (x == 1 && y == 4)
		PORTD = 0x40; // bit-7 - D

	if (x == 2 && y == 1)
		PORTD = 0x01; // bit-1 - D
	if (x == 2 && y == 2)
		PORTD = 0x04; // bit-3 - D
	if (x == 2 && y == 3)
		PORTD = 0x80; // bit-8 - D
	if (x == 2 && y == 4)
		PORTB = 0x01; // bit-1 - B

	if (x == 3 && y == 1)
		PORTD = 0x10; // bit-5 - D
	if (x == 3 && y == 2)
		PORTD = 0x20; // bit-6 - D
	if (x == 3 && y == 3)
		PORTC = 0x04; // bit-3 - C
	if (x == 3 && y == 4)
		PORTB = 0x08; // bit-4 - B

	if (x == 4 && y == 1)
		PORTC = 0x20; // bit-6 - C
	if (x == 4 && y == 2)
		PORTC = 0x02; // bit-2 - C
	if (x == 4 && y == 3)
		PORTC = 0x08; // bit-4 - C
	if (x == 4 && y == 4)
		PORTB = 0x04; // bit-3 - B

	if (x == 5 && y == 1)
		PORTB = 0x10; // bit-5 - B
	if (x == 5 && y == 2)
		PORTB = 0x20; // bit-6 - B
	if (x == 5 && y == 3)
		PORTC = 0x10; // bit-5 - C
	if (x == 5 && y == 4)
		PORTB = 0x02; // bit-2 - B
	sleep(sleepTime);
}

void lightOff() {
	PORTB = 0x00;
	PORTC = 0x00;
	PORTD = 0x00;
}

void sleep(uint8_t millisec) {
	while (millisec)

	{
		_delay_ms(1);/* 1 ms delay */
		millisec--;
	}
}
